//+------------------------------------------------------------------+
//|                                                opencl_program.cl |
//|                                  Copyright 2021, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
//--- by default some GPU doesn't support doubles
//--- cl_khr_fp64 directive is used to enable work with doubles
#pragma OPENCL EXTENSION cl_khr_fp64 : enable
//+------------------------------------------------------------------+
//| Activation functions                                             |
//+------------------------------------------------------------------+
//| Linear activation function                                       |
//| Parameter 'value' Weighted sum of initial data                   |
//|           'a' defines the angle of inclination of the line       |
//|           'b' - the vertical offset of the line                  |
//+------------------------------------------------------------------+
double LineActivation(double value, double a, double b)
  {
   return (a * value + b);
  }
//+------------------------------------------------------------------+
//| Sigmoid activation function                                      |
//| Parameter 'value' Weighted sum of initial data                   |
//|           'a' stretches the range of values of the function      |
//|               from '0' to 'a'                                    |
//|           'b' shifts the resulting value                         |
//+------------------------------------------------------------------+
double SigmoidActivation(double value, double a, double b)
  {
   return (a / (1 + exp(-value)) - b);
  }
//---
double4 SigmoidActivation4(double4 value, double a, double b)
  {
   return (a / (1 + exp(-value)) - b);
  }
//+------------------------------------------------------------------+
//| Derivative of Sigmoid activation function                        |
//| Parameter 'value' current point (result of feed forward).        |
//|           'a' stretches the range of values of the function from |
//|               '0' to 'a'                                         |
//|           'b' shifts the resulting value                         |
//+------------------------------------------------------------------+
double4 SigmoidDerivative(double4 value, double a, double b)
  {
   if(a == 0)
      return (double4)(0);
//---
   double4 z = max(min((double4)a, value + b), (double4)(0));
   return (z * (1 - z / a));
  }
//+------------------------------------------------------------------+
//| TANH activation function                                         |
//| Parameter 'value' Weighted sum of initial data                   |
//+------------------------------------------------------------------+
double TanhActivation(double value)
  {
   return tanh(value);
  }
//+------------------------------------------------------------------+
//| Derivative of TANH activation function                           |
//| Parameter 'value' current point (result of feed forward).        |
//+------------------------------------------------------------------+
double4 TanhDerivative(double4 value)
  {
   return (1 - pow(value, 2));
  }
//+------------------------------------------------------------------+
//| LReLU activation function                                        |
//| Parameter 'value' current point (result of feed forward).        |
//|           'a' leak parameter                                     |
//+------------------------------------------------------------------+
double LReLUActivation(double value, double a)
  {
   return (value > 0 ? value : a * value);
  }
//+------------------------------------------------------------------+
//| Derivative of LReLU activation function                           |
//| Parameter 'value' current point (result of feed forward).        |
//|           'a' leak parameter                                     |
//+------------------------------------------------------------------+
double4 LReLUDerivative(double4 value, double a)
  {
   double4 result = (double4)((value.s0 > 0 ? 1 : a), (value.s1 > 0 ? 1 : a), (value.s2 > 0 ? 1 : a), (value.s3 > 0 ? 1 : a));
   return (result);
  }
//+------------------------------------------------------------------+
//| Swish activation function                                        |
//| Parameter 'value' Weighted sum of initial data                   |
//|           'b' affects the nonlinearity of the function           |
//+------------------------------------------------------------------+
double SwishActivation(double value, double b)
  {
   return value * SigmoidActivation(b * value, 1, 0);
  }
//+------------------------------------------------------------------+
//| Derivative of Swish activation function                          |
//| Parameter 'value' current point (result of feed forward).        |
//|           'value_input' Weighted sum of initial data             |
//|           'b' affects the nonlinearity of the function           |
//+------------------------------------------------------------------+
double4 SwishDerivative(double4 value, double4 input_value, double b)
  {
//---
   double4 by = b * value;
   return (by + SigmoidActivation4(b * input_value, 1, 0) * (1 - by));
  }
//+------------------------------------------------------------------+
//| Transfer 4 elements of double vector to double4                  |
//| Parameter 'array' source array of data                           |
//|           'start' first position to copy                         |
//|           'step' step between elements to copy                   |
//|           'size' Size of source array                            |
//|         'shift' Shift in source array to the 1-st copied element |
//+------------------------------------------------------------------+
double4 ToVect4(__global double *array, int start, int step, int size, int shift)
  {
   double4 result = (double4)(0, 0, 0, 0);
   step = max(1, step);
   int st = start * step + shift;
   if(st < size)
     {
      int k = (size - shift) % step;
      k = (size - shift - k) / step - start + (k > 0 ? 1 : 0);
      switch(k)
        {
         case 0:
            break;
         case  1:
            result = (double4)(array[st], 0, 0, 0);
            break;
         case  2:
            result = (double4)(array[st], array[st + step], 0, 0);
            break;
         case  3:
            result = (double4)(array[st], array[st + step], array[st + 2 * step], 0);
            break;
         default:
            result = (double4)(array[st], array[st + step], array[st + 2 * step], array[st + 3 * step]);
            break;
        }
     }
   return result;
  }
//+------------------------------------------------------------------+
//| Derivative of SoftMax activation function                        |
//| Parameter 'outputs' vector of previous FF outputs data           |
//|           'gradients' vector of gradients                        |
//|           'shift' shift to first position in vector              |
//|           'outputs_total' size of outputs vector                 |
//+------------------------------------------------------------------+
double4 SoftMaxDerivative(__global double *outputs, __global double *gradient, int shift, int outputs_total)
  {
   double4 result = (double4)0;
   for(int j = 0; j < outputs_total; j++)
      result += (double4)(outputs[j] * gradient[j]) * ((double4)(shift == j, (shift + 1) == j, (shift + 2) == j, (shift + 3) == j) - ToVect4(outputs, shift, 1, outputs_total, 0));
   return result;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Max4(double4 vect, double value)
  {
   double result = fmax(vect.s0, value);
   result = fmax(vect.s1, result);
   result = fmax(vect.s2, result);
   return(fmax(vect.s3, result));
  }
//+------------------------------------------------------------------+
//| Transfer double4 to 4 elements of double vector                  |
//| Parameter 'array' target array of data                           |
//|           'value' source double4 vector                          |
//|           'start' first position to copy in target array         |
//|           'step' step between elements in target array           |
//|           'size' Size of target array                            |
//|         'shift' Shift in target array to the 1-st copied element |
//+------------------------------------------------------------------+
void D4ToArray(__global double *array, double4 value, int start, int step, int size, int shift)
  {
   step = max(1, step);
   int st = start * step + shift;
   if(st < size)
     {
      int k = (size - shift) % step;
      k = (size - shift - k) / step - start + (k > 0 ? 1 : 0);
      switch(k)
        {
         case  3:
            array[st + 2 * step] = value.s2;
         case  2:
            array[st + step] = value.s1;
         case  1:
            array[st] = value.s0;
            break;
         default:
            array[st + 3 * step] = value.s3;
            array[st + 2 * step] = value.s2;
            array[st + step] = value.s1;
            array[st] = value.s0;
            break;
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//| Kernel Feed Forward of perceptron                                |
//| Parameter 'inputs' vector of inputs data                         |
//|           'weights' matrix of weights                            |
//|           'sums' vector of multiploer inputs to weights          |
//|           'outputs' output data after activation function        |
//|           'inputs_total' size of inputs vector                   |
//|           'activation' type of activation function               |
//|           'act_param_a' 1st parameter of activation function     |
//|           'act_param_b' 2nd parameter of activation function     |
//+------------------------------------------------------------------+
__kernel void PerceptronFeedForward(__global double *inputs,
                                    __global double *weights,
                                    __global double *sums,
                                    __global double *outputs,
                                    int inputs_total, int activation,
                                    double act_param_a,
                                    double act_param_b)
  {
   const int n = get_global_id(0);
   const int weights_total = get_global_size(0) * (inputs_total + 1);
   int shift = n * (inputs_total + 1);
   double s = weights[shift + inputs_total];
   for(int i = 0; i < inputs_total; i += 4)
      s += dot(ToVect4(inputs, i, 1, inputs_total, 0), ToVect4(weights, i, 1, weights_total, shift));
   switch(activation)
     {
      case 0:
         outputs[n] = LineActivation(s, act_param_a, act_param_b);
         break;
      case 1:
         outputs[n] = SigmoidActivation(s, act_param_a, act_param_b);
         break;
      case 2:
         outputs[n] = TanhActivation(s);
         break;
      case 3:
         outputs[n] = LReLUActivation(s, act_param_a);
         break;
      case 4:
         outputs[n] = exp(s);
         break;
      case 5:
         sums[n] = s;
         outputs[n] = SwishActivation(s, act_param_a);
         break;
      default:
         outputs[n] = s;
         break;
     }
  }
//+------------------------------------------------------------------+
//| Kernel of divide vector to constant value                        |
//| Parameter 'inputs' vector of inputs data                         |
//|           'outputs' vector of output data                        |
//|           'inputs_total' size of inputs vector                   |
//|           'const_value' constant value                           |
//+------------------------------------------------------------------+
__kernel void Normalize(__global double *inputs,
                        __global double *outputs,
                        int inputs_total,
                        double const_value)
  {
   const int n = get_global_id(0);
   int shift = n * 4;
   double4 result = ToVect4(inputs, shift, 1, inputs_total, 0) / (double4)(const_value);
   D4ToArray(outputs, result, shift, 1, inputs_total, 0);
  }
//+------------------------------------------------------------------+
//| Kernel of calculation output gradients                           |
//| Parameter 'target' vector of target data                         |
//|           'outputs' vector of previous FF outputs data           |
//|           'gradients' vector of gradients                        |
//|           'outputs_total' size of outputs vector                 |
//+------------------------------------------------------------------+
__kernel void CalcOutputGradient(__global double *target,
                                 __global double *outputs,
                                 __global double *gradients,
                                 int outputs_total)
  {
   const int n = get_global_id(0);
   double4 result = ToVect4(target, n * 4, 1, outputs_total, 0) - ToVect4(outputs, n * 4, 1, outputs_total, 0);
   D4ToArray(gradients, result, n * 4, 1, outputs_total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void DeActivateGradient(__global double *sums,
                                 __global double *outputs,
                                 __global double *gradients,
                                 int outputs_total, int activation,
                                 double act_param_a,
                                 double act_param_b)
  {
   const int n = get_global_id(0);
   int shift = n * 4;
   double4 grad = ToVect4(gradients, shift, 1, outputs_total, 0);
   double4 out = ToVect4(outputs, shift, 1, outputs_total, 0);
   double4 s = ToVect4(sums, shift, 1, outputs_total, 0);
   switch(activation)
     {
      case 0:
         grad = grad * act_param_a;
         break;
      case 1:
         grad = grad * SigmoidDerivative(out, act_param_a, act_param_b);
         break;
      case 2:
         grad = grad * TanhDerivative(out);
         break;
      case 3:
         grad = grad * LReLUDerivative(out, act_param_a);
         break;
      case 4:
         grad = SoftMaxDerivative(outputs, sums, shift, outputs_total);
         break;
      case 5:
         grad = grad * SwishDerivative(out, s, act_param_a);
         break;
      default:
         break;
     }
   D4ToArray(gradients, grad, shift, 1, outputs_total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void CalcHiddenGradient(__global double *gradient_inputs,
                                 __global double *weights,
                                 __global double *gradients,
                                 int outputs_total)
  {
   const int n = get_global_id(0);
   const int inputs_total = get_global_size(0);
   int weights_total = (inputs_total + 1) * outputs_total;
//---
   double grad = 0;
   for(int o = 0; o < outputs_total; o += 4)
      grad += dot(ToVect4(gradients, o, 1, outputs_total, 0), ToVect4(weights, o, (inputs_total + 1), weights_total, n));
   gradient_inputs[n] = grad;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void CalcDeltaWeights(__global double *inputs,
                               __global double *delta_weights,
                               __global double *gradients,
                               int inputs_total)
  {
   const int n = get_global_id(0);
   const int outputs_total = get_global_size(0);
   int weights_total = (inputs_total + 1) * outputs_total;
//---
   double grad = gradients[n];
   int shift = n * (inputs_total + 1);
   for(int i = 0; i < inputs_total; i += 4)
     {
      double4 delta = ToVect4(inputs, i, 1, inputs_total, 0) * grad + ToVect4(delta_weights, i, 1, weights_total, shift);
      D4ToArray(delta_weights, delta, i, 1, weights_total, shift);
     }
   delta_weights[shift + inputs_total] += grad;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void SGDUpdate(__global double *delta_weights,
                        __global double *weights,
                        int total,
                        int batch_size,
                        double learningRate,
                        double Lambda1,
                        double Lambda2
                       )
  {
   int start = 4 * get_global_id(0);
   double4 delta4 = ToVect4(delta_weights, start, 1, total, 0);
   double4 weights4 = ToVect4(weights, start, 1, total, 0);
   double lr = learningRate / ((double)batch_size);
   weights4 -= (double4)(Lambda1) + Lambda2 * weights4;
   weights4 += (double4)(lr) * delta4;
   D4ToArray(weights, weights4, start, 1, total, 0);
   D4ToArray(delta_weights, (double4)(0), start, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void MomentumUpdate(__global double *delta_weights,
                             __global double *weights,
                             __global double *momentum,
                             int total, int batch_size,
                             double learningRate,
                             double beta,
                             double Lambda1, double Lambda2)
  {
   int start = 4 * get_global_id(0);
//---
   double4 delta4 = ToVect4(delta_weights, start, 1, total, 0) / ((double4)(batch_size));
   double4 weights4 = ToVect4(weights, start, 1, total, 0);
   double4 momentum4 = ToVect4(momentum, start, 1, total, 0);
   weights4 -= (double4)(Lambda1) + Lambda2 * weights4;
   momentum4 = (double4)(learningRate) * delta4 + (double4)(beta) * momentum4;
   weights4 += momentum4;
   D4ToArray(weights, weights4, start, 1, total, 0);
   D4ToArray(momentum, momentum4, start, 1, total, 0);
   D4ToArray(delta_weights, (double4)(0), start, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void AdaGradUpdate(__global double *delta_weights,
                            __global double *weights,
                            __global double *momentum,
                            int total, int batch_size,
                            double learningRate,
                            double Lambda1, double Lambda2)
  {
   int start = 4 * get_global_id(0);
//---
   double4 delta4 = ToVect4(delta_weights, start, 1, total, 0) / ((double4)(batch_size));
   double4 weights4 = ToVect4(weights, start, 1, total, 0);
   double4 momentum4 = ToVect4(momentum, start, 1, total, 0);
//---
   weights4 -= (double4)(Lambda1) + Lambda2 * weights4;
   double4 G = momentum4 + pow(delta4, 2);
   weights4 += learningRate / sqrt(G + 1.0e-10);
   D4ToArray(weights, weights4, start, 1, total, 0);
   D4ToArray(momentum, momentum4, start, 1, total, 0);
   D4ToArray(delta_weights, (double4)(0), start, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void RMSPropUpdate(__global double *delta_weights,
                            __global double *weights,
                            __global double *momentum,
                            int total, int batch_size,
                            double learningRate,
                            double beta,
                            double Lambda1, double Lambda2)
  {
   int start = 4 * get_global_id(0);
//---
   double4 delta4 = ToVect4(delta_weights, start, 1, total, 0) / ((double4)(batch_size));
   double4 weights4 = ToVect4(weights, start, 1, total, 0);
   double4 momentum4 = ToVect4(momentum, start, 1, total, 0);
//---
   weights4 -= (double4)(Lambda1) + Lambda2 * weights4;
   double4 G = beta * momentum4 + (1 - beta) * pow(delta4, 2);
   weights4 += learningRate / (sqrt(G) + 1.0e-10) * delta4;
   D4ToArray(weights, weights4, start, 1, total, 0);
   D4ToArray(momentum, momentum4, start, 1, total, 0);
   D4ToArray(delta_weights, (double4)(0), start, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void AdaDeltaUpdate(__global double *delta_weights,
                             __global double *weights,
                             __global double *momentumW,
                             __global double *momentumG,
                             int total, int batch_size,
                             double beta1, double beta2,
                             double Lambda1, double Lambda2)
  {
   int start = 4 * get_global_id(0);
//---
   double4 delta4 = ToVect4(delta_weights, start, 1, total, 0) / ((double4)(batch_size));
   double4 weights4 = ToVect4(weights, start, 1, total, 0);
   double4 momentumW4 = ToVect4(momentumW, start, 1, total, 0);
   double4 momentumG4 = ToVect4(momentumG, start, 1, total, 0);
//---
   weights4 -= (double4)(Lambda1) + Lambda2 * weights4;
   double4 W = beta1 * momentumW4 + (1 - beta1) * pow(weights4, 2);
   double4 G = beta2 * momentumG4 + (1 - beta2) * pow(delta4, 2);
   weights4 += sqrt(W) / (sqrt(G) + 1.0e-10) * delta4;
   D4ToArray(weights, weights4, start, 1, total, 0);
   D4ToArray(momentumW, momentumW4, start, 1, total, 0);
   D4ToArray(momentumG, momentumG4, start, 1, total, 0);
   D4ToArray(delta_weights, (double4)(0), start, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void AdamUpdate(__global double *delta_weights,
                         __global double *weights,
                         __global double *momentumM,
                         __global double *momentumV,
                         int total, int batch_size,
                         double learningRate,
                         double beta1, double beta2,
                         double Lambda1, double Lambda2)
  {
   int start = 4 * get_global_id(0);
//---
   double4 delta4 = ToVect4(delta_weights, start, 1, total, 0) / ((double4)(batch_size));
   double4 weights4 = ToVect4(weights, start, 1, total, 0);
   double4 momentumM4 = ToVect4(momentumM, start, 1, total, 0);
   double4 momentumV4 = ToVect4(momentumV, start, 1, total, 0);
//---
   double4 M = beta1 * momentumM4 + (1 - beta1) * delta4;
   double4 V = beta2 * momentumV4 + (1 - beta2) * pow(delta4, 2);
   double4 m = M / (1 - beta1);
   double4 v = V / (1 - beta2);
   weights4 -= (double4)(Lambda1) + Lambda2 * weights4;
   weights4 += learningRate * m / (sqrt(v) + 1.0e-10);
   D4ToArray(weights, weights4, start, 1, total, 0);
   D4ToArray(momentumM, momentumM4, start, 1, total, 0);
   D4ToArray(momentumV, momentumV4, start, 1, total, 0);
   D4ToArray(delta_weights, (double4)(0), start, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                               |
//|  'inputs'                         |
//|           'weights'                                  |
//|           'sums'       |
//|           'outputs'                             |
//|           'inputs_total'              |
//|           'window'                |
//|           'step'                              |
//|           'window_out'                         |
//|           'activation'                        |
//|           'act_param_a' 1-              |
//|           'act_param_b' 2-             |
//+------------------------------------------------------------------+
__kernel void ConvolutionFeedForward(__global double *inputs,
                                     __global double *weights,
                                     __global double *sums,
                                     __global double *outputs,
                                     int inputs_total,
                                     int window,
                                     int step,
                                     int window_out,
                                     int activation,
                                     double act_param_a,
                                     double act_param_b,
                                     int transposed_out)
  {
   const int n = get_global_id(0);
   const int neurons = get_global_size(0);
   const int weights_total = (window + 1) * window_out;
   int shift = n * step;
   for(int w = 0; w < window_out; w++)
     {
      int out = (transposed_out == 1 ? w + n * window_out : w * neurons + n);
      int shift_weights = w * (window + 1) ;
      if((shift_weights + window) >= weights_total)
         break;
      double s = weights[shift_weights + window];
      for(int i = 0; i < window; i += 4)
         s += dot(ToVect4(inputs, i, 1, inputs_total, shift),
                  ToVect4(weights, i, 1, weights_total, shift_weights));
      switch(activation)
        {
         case 0:
            outputs[out] = LineActivation(s, act_param_a, act_param_b);
            break;
         case 1:
            outputs[out] = SigmoidActivation(s, act_param_a, act_param_b);
            break;
         case 2:
            outputs[out] = TanhActivation(s);
            break;
         case 3:
            outputs[out] = LReLUActivation(s, act_param_a);
            break;
         case 4:
            outputs[out] = exp(s);
            break;
         case 5:
            sums[out] = s;
            outputs[out] = SwishActivation(s, act_param_a);
            break;
         default:
            outputs[out] = s;
            break;
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void ConvolutionCalcHiddenGradient(__global double *gradient_inputs,
      __global double *weights,
      __global double *gradients,
      int outputs_total,
      int window,
      int step,
      int window_out,
      int neurons,
      int transposed_out)
  {
   const int n = get_global_id(0);
   const int inputs_total = get_global_size(0);
   int weights_total = (window + 1) * window_out;
//---
   double grad = 0;
   int start = n - window + step;
   start = max((start - start % step) / step, 0);
   int stop = min((n - n % step) / step + 1, neurons);
   for(int w = 0; w < window_out; w++)
     {
      int shift_w = w * (window + 1);
      if((shift_w + window) > weights_total)
         break;
      for(int o = start; o < stop; o ++)
        {
         int shift_wl = (stop - o - 1) * step + n % step + shift_w;
         int shift_gl = (transposed_out == 1 ? w + o * window_out : w * neurons + o);
         grad += gradients[shift_gl] * weights[shift_wl];
        }
     }
   gradient_inputs[n] = grad;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void ConcolutionCalcDeltaWeights(__global double *inputs,
      __global double *delta_weights,
      __global double *gradients,
      int inputs_total,
      int step,
      int neurons,
      int transposed_out)
  {
   const int inp_w = get_global_id(0);
   const int w = get_global_id(1);
   const int window = get_global_size(0) - 1;
   const int window_out = get_global_size(1);
//   int weights_total = (window + 1) * window_out;
//---
   int shift_delt = w * (window + 1) + inp_w;
   double value = 0;
   if(inp_w == window)
     {
      for(int n = 0; n < neurons; n ++)
         value += gradients[transposed_out == 1 ? w + n*window_out : w * neurons + n];
     }
   else
      for(int n = 0; n < neurons; n ++)
        {
         int shift_inp = n * step + inp_w;
         if(shift_inp >= inputs_total)
            break;
         value += inputs[shift_inp] * gradients[transposed_out == 1 ? w + n*window_out : w * neurons + n];
        }
   delta_weights[shift_delt] += value;
  }
//+------------------------------------------------------------------+
//|                            |
//|  'inputs'                         |
//|           'outputs'                             |
//|           'inputs_total'              |
//|           'input_neurons'   1-  . |
//|           'window'                |
//|           'step'                              |
//|           'window_out'                         |
//|           'activation'                        |
//+------------------------------------------------------------------+
__kernel void ProofFeedForward(__global double *inputs,
                               __global double *outputs,
                               int inputs_total,
                               int input_neurons,
                               int window,
                               int step,
                               int window_out,
                               int activation)
  {
   const int n = get_global_id(0);
   const int neurons = get_global_size(0);
   int shift = n * step;
   for(int w = 0; w < window_out; w++)
     {
      int out = w * neurons + n;
      int shift_inp = w * input_neurons;
      double s = 0;
      double k = 1.0 / (double)window;
      double4 k4 = (double4)(k);
      for(int i = 0; i < window; i += 4)
         switch(activation)
           {
            case 6:
               s += dot(ToVect4(inputs, i, 1, min(shift_inp + input_neurons, inputs_total), shift_inp + shift),
                        k4);
               break;
            case 7:
               s = Max4(ToVect4(inputs, i, 1, min(shift_inp + input_neurons, inputs_total), shift_inp + shift), s);
               break;
            default:
               break;
           }
      outputs[out] = s;
     }
  }
//+------------------------------------------------------------------+
//|                          |
//|  'inputs'                         |
//|           'gradient_inputs'   |
//|           'outputs'                             |
//|           'gradients'                |
//|           'inputs_total'              |
//|           'outputs_total'                |
//|           'window'                |
//|           'step'                              |
//|           'window_out'                         |
//|           'neurons'   1-        |
//|           'activation'                        |
//+------------------------------------------------------------------+
__kernel void ProofCalcHiddenGradient(__global double *inputs,
                                      __global double *gradient_inputs,
                                      __global double *outputs,
                                      __global double *gradients,
                                      int inputs_total,
                                      int outputs_total,
                                      int window,
                                      int step,
                                      int window_out,
                                      int neurons,
                                      int activation)
  {
   const int n = get_global_id(0);
   const int input_neurons = get_global_size(0);
//---
   int start = n - window + step;
   start = max((start - start % step) / step, 0);
   int stop = min((n - n % step) / step + 1, neurons);
   for(int w = 0; w < window_out; w++)
     {
      double grad = 0;
      int shift_inp = w * input_neurons + n;
      if(shift_inp >= inputs_total)
         break;
      double inp = inputs[shift_inp];
      for(int o = start; o < stop; o ++)
        {
         int shift_g = w * neurons + o;
         if(shift_g >= outputs_total)
            break;
         switch(activation)
           {
            case 6:
               grad += gradients[shift_g] / (double)window;
               break;
            case 7:
               grad += (outputs[shift_g] == inp ? gradients[shift_g] : 0);
               break;
            default:
               break;
           }
        }
      gradient_inputs[shift_inp] = grad;
     }
  }
//+------------------------------------------------------------------+
//|    LSTM                                 |
//|  'forgetgate'                              |
//|           'inputgate'                                |
//|           'outputgate'                           |
//|           'newcontent'                               |
//|           'memory'                                    |
//|           'hiddenstate'                    |
//|           'outputs_total'        |
//+------------------------------------------------------------------+
__kernel void LSTMFeedForward(__global double *forgetgate,
                              __global double *inputgate,
                              __global double *outputgate,
                              __global double *newcontent,
                              __global double *memory,
                              __global double *hiddenstate,
                              int outputs_total)
  {
   const int n = get_global_id(0);
   const int shift = n * 4;
   double4 fg = ToVect4(forgetgate, shift, 1, outputs_total, 0);
   double4 ig = ToVect4(inputgate, shift, 1, outputs_total, 0);
   double4 og = ToVect4(outputgate, shift, 1, outputs_total, 0);
   double4 nc = ToVect4(newcontent, shift, 1, outputs_total, 0);
   double4 mem = ToVect4(memory, shift, 1, outputs_total, 0);
//---
   double4 temp = mem * fg;
   temp += ig * nc;
   D4ToArray(memory, temp, shift, 1, outputs_total, 0);
   temp = tanh(temp) * og;
   D4ToArray(hiddenstate, temp, shift, 1, outputs_total, 0);
  }
//+------------------------------------------------------------------+
//|    LSTM                               |
//|  'outputs'                             |
//|           'gradients'                |
//|           'inputgate'                                |
//|           'outputgate'                           |
//|           'newcontent'                               |
//|           'memory'                                    |
//|           'fg_gradients'                    |
//|           'ig_gradients'                      |
//|           'og_gradients'                  |
//|           'nc_gradients'                   |
//|           'outputs_total'                |
//+------------------------------------------------------------------+
__kernel void LSTMCalcHiddenGradient(__global double *outputs,
                                     __global double *gradients,
                                     __global double *inputgate,
                                     __global double *outputgate,
                                     __global double *newcontent,
                                     __global double *memory,
                                     __global double *fg_gradients,
                                     __global double *ig_gradients,
                                     __global double *og_gradients,
                                     __global double *nc_gradients,
                                     int outputs_total)
  {
   const int n = get_global_id(0);
   int shift = n * 4;
//---
   double4 out = ToVect4(outputs, shift, 1, outputs_total, 0);
   double4 grad = ToVect4(gradients, shift, 1, outputs_total, 0);
   double4 ig = ToVect4(inputgate, shift, 1, outputs_total, 0);
   double4 og = ToVect4(outputgate, shift, 1, outputs_total, 0);
   double4 nc = ToVect4(newcontent, shift, 1, outputs_total, 0);
   double4 mem = ToVect4(memory, shift, 1, outputs_total, 0);
//---
   double4 m = out / (og + 1.0e-8);
//--- OutputGate 
   double4 temp = grad * m;
   D4ToArray(og_gradients, temp, shift, 1, outputs_total, 0);
//---   c   TANH
   grad = grad * og * (1 - pow(m, 2));
//--- InputGate 
   temp = grad * nc;
   D4ToArray(ig_gradients, temp, shift, 1, outputs_total, 0);
//--- NewContent 
   temp = grad * ig;
   D4ToArray(nc_gradients, temp, shift, 1, outputs_total, 0);
//--- ForgetGates 
   temp = grad * mem;
   D4ToArray(fg_gradients, temp, shift, 1, outputs_total, 0);
  }
//+------------------------------------------------------------------+
//|      Self-Attention |
//|  'querys'                                 |
//|           'keys'                                     |
//|           'values'                                 |
//|           'scores'               |
//|           'inputs'                           |
//|           'outputs'                             |
//|           'window'                |
//|           'key_size'            |
//+------------------------------------------------------------------+
__kernel void AttentionFeedForward(__global double *querys,
                                   __global double *keys,
                                   __global double *scores,
                                   __global double *values,
                                   __global double *outputs,
                                   int window,
                                   int key_size,
                                   int mask)
  {
   const int q = get_global_id(0);
   const int units = get_global_size(0);
   const int h = get_global_id(1);
   const int heads = get_global_size(1);
   int shift_query = key_size * (q * heads + h);
   int shift_scores = units * (q * heads + h);
   double summ = 0;
   for(int s = 0; s < units; s++)
     {
      double score = 0;
      if(mask > 0 && s > q)
        {
         scores[shift_scores + s] = score;
         continue;
        }
      int shift_key = key_size * (s * heads + h);
      for(int k = 0; k < key_size; k ++)
         score += querys[shift_query + k] * keys[shift_key + k];
      score = exp(score / sqrt((double)key_size));
      summ += score;
      scores[shift_scores + s] = score;
     }
   for(int s = 0; s < units; s++)
      scores[shift_scores + s] /= summ;
//---
   shift_query = window * (q * heads + h);
   for(int i = 0; i < window; i++)
     {
      double query = 0;
      for(int v = 0; v < units; v++)
         query += values[window * (v * heads + h) + i] * scores[shift_scores + v];
      outputs[shift_query + i] = query;
     }
  }
//+------------------------------------------------------------------+
//|      Self-Attention      |
//|      Score                |
//|  'values'                                 |
//|           'values_grad'          |
//|           'scores'               |
//|           'scores_grad'   .     |
//|           'outputs'                             |
//|           'outputs_grad'      |
//|           'window'                |
//+------------------------------------------------------------------+
__kernel void AttentionCalcScoreGradient(__global double *scores,
      __global double *scores_grad,
      __global double *values,
      __global double *values_grad,
      __global double *outputs_grad,
      __global double *scores_temp,
      int window)
  {
   const int q = get_global_id(0);
   const int units = get_global_size(0);
   const int h = get_global_id(1);
   const int heads = get_global_size(1);
   int shift_value = window * (q * heads + h);
   int shift_score = units * (q * heads + h);
//---    Values
   for(int i = 0; i < window; i ++)
     {
      double grad = 0;
      for(int g = 0; g < units; g++)
         grad += scores[units * (g * heads + h) + q] * outputs_grad[window * (g * heads + h) + i];
      values_grad[shift_value + i] = grad;
     }
//---    Score
   for(int k = 0; k < units; k++)
     {
      double grad = 0;
      for(int i = 0; i < window; i++)
         grad += outputs_grad[shift_value + i] * values[window * (k * heads + h) + i];
      scores_temp[shift_score + k] = grad;
     }
//---    Softmax
   for(int k = 0; k < units; k++)
     {
      double grad = 0;
      double score = scores[shift_score + k];
      for(int i = 0; i < units; i++)
         grad += scores[shift_score + i] * ((int)(i == k) - score) * scores_temp[shift_score + i];
      scores_grad[shift_score + k] = grad;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void AttentionCalcHiddenGradient(__global double *querys,
      __global double *querys_grad,
      __global double *keys,
      __global double *keys_grad,
      __global double *scores_grad,
      int key_size)
  {
   const int q = get_global_id(0);
   const int units = get_global_size(0);
   const int h = get_global_id(1);
   const int heads = get_global_size(1);
   int shift_query = key_size * (q * heads + h);
   int shift_score = units * (q * heads + h);
//---    Querys  Keys
   const double k = 1 / sqrt((double)key_size);
//---
   for(int i = 0; i < key_size; i++)
     {
      double grad_q = 0;
      double grad_k = 0;
      for(int s = 0; s < units; s++)
        {
         grad_q += keys[key_size * (s * heads + h) + i] * scores_grad[shift_score + s];
         grad_k += querys[key_size * (s * heads + h) + i] * scores_grad[units * (s * heads + h) + q];
        }
      querys_grad[shift_query + i] = grad_q * k;
      keys_grad[shift_query + i] = grad_k * k;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void GPTFeedForward(__global double *querys,
                             __global double *keys,
                             __global double *scores,
                             __global double *values,
                             __global double *outputs,
                             int key_size,
                             int units)
  {
   const int h = get_global_id(0);
   const int heads = get_global_size(0);
   int shift_query = key_size * h;
   int shift_scores = units * h;
   double summ = 0;
   for(int s = 0; s < units; s++)
     {
      double score = 0;
      int shift_key = key_size * (s * heads + h);
      for(int k = 0; k < key_size; k ++)
         score += querys[shift_query + k] * keys[shift_key + k];
      score = exp(score / sqrt((double)key_size));
      summ += score;
      scores[shift_scores + s] = score;
     }
   for(int s = 0; s < units; s++)
      scores[shift_scores + s] /= summ;
//---
   shift_query = key_size * h;
   for(int i = 0; i < key_size; i++)
     {
      double query = 0;
      for(int v = 0; v < units; v++)
         query += values[key_size * (v * heads + h) + i] * scores[shift_scores + v];
      outputs[shift_query + i] = query;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void GPTCalcScoreGradient(__global double *scores,
                                   __global double *scores_grad,
                                   __global double *values,
                                   __global double *values_grad,
                                   __global double *outputs_grad,
                                   __global double *scores_temp,
                                   int window,
                                   int units,
                                   int current)
  {
   const int h = get_global_id(0);
   const int heads = get_global_size(0);
   int shift_value = window * (2 * heads + h);
   int shift_score = units * h;
//---    Values
   for(int i = 0; i < window; i ++)
      values_grad[shift_value + i] = scores[units * h + current] * outputs_grad[window * h + i];
//---    Score
   for(int k = 0; k < units; k++)
     {
      double grad = 0;
      for(int i = 0; i < window; i++)
         grad += outputs_grad[shift_value + i] * values[window * (k * heads + h) + i];
      scores_temp[shift_score + k] = grad;
     }
//---    Softmax
   for(int k = 0; k < units; k++)
     {
      double grad = 0;
      double score = scores[shift_score + k];
      for(int i = 0; i < units; i++)
         grad += scores[shift_score + i] * ((int)(i == k) - score) * scores_temp[shift_score + i];
      scores_grad[shift_score + k] = grad;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void GPTCalcHiddenGradient(__global double *querys,
                                    __global double *querys_grad,
                                    __global double *keys,
                                    __global double *scores_grad,
                                    int key_size,
                                    int units,
                                    int current)
  {
   const int h = get_global_id(0);
   const int heads = get_global_size(0);
   int shift_query = key_size * h;
   int shift_key = key_size * (heads + h);
   int shift_score = units * h;
//---    Querys  Keys
   const double k = 1 / sqrt((double)key_size);
//---
   for(int i = 0; i < key_size; i++)
     {
      double grad_q = 0;
      double grad_k = 0;
      for(int s = 0; s < units; s++)
        {
         grad_q += keys[key_size * (s * heads + h) + i] * scores_grad[shift_score + s];
         if(s == current)
            grad_k += querys[key_size * h + i] * scores_grad[units * h + current];
        }
      querys_grad[shift_query + i] = grad_q * k;
      querys_grad[shift_key + i] = grad_k * k;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void BatchNormFeedForward(__global double *inputs,
                                   __global double *options,
                                   __global double *weights,
                                   __global double *output,
                                   int batch,
                                   int total)
  {
   int n = get_global_id(0);
   if(batch <= 1)
     {
      D4ToArray(output, ToVect4(inputs, n * 4, 1, total, 0), n * 4, 1, total, 0);
      return;
     }
   int shift = n * 4;
   int shift_options = n * 3 * 4;
   int shift_weights = n * 2 * 4;
//---
   double4 inp = ToVect4(inputs, shift, 1, total, 0);
   double4 mean = ToVect4(options, shift, 3, total * 3, 0) * ((double)batch - 1.0) + inp ;
   if(options[shift_options ] > 0 && options[shift_options + 1] > 0)
      mean /= (double4)batch;
   double4 delt = inp - mean;
   double4 variance = ToVect4(options, shift, 3, total * 3, 1) * ((double)batch - 1.0) + pow(delt, 2);
   if(options[shift_options + 1] > 0)
      variance /= (double4)batch;
   double4 nx = delt / sqrt(variance + 1e-6);
//---
   if(weights[shift_weights] == 0)
      D4ToArray(weights, (double4)1.0, shift, 2, total * 2, 0);
//---
   double4 res = ToVect4(weights, shift, 2, total * 2, 0) * nx + ToVect4(weights, shift, 2, total * 2, 1);
//---
   D4ToArray(options, mean, shift, 3, total * 3, 0);
   D4ToArray(options, variance, shift, 3, total * 3, 1);
   D4ToArray(options, nx, shift, 3, total * 3, 2);
   D4ToArray(output, res, shift, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void BatchNormCalcHiddenGradient(__global double *options,
      __global double *gradient,
      __global double *inputs,
      __global double *gradient_inputs,
      __global double *weights,
      int batch,
      int total
                                         )
  {
   int n = get_global_id(0);
   int shift = n * 4;
   if(batch <= 1)
     {
      D4ToArray(gradient_inputs, ToVect4(gradient, shift, 1, total, 0), shift, 1, total, 0);
      return;
     }
//---
   double4 inp = ToVect4(inputs, shift, 1, total, 0);
   double4 gnx = ToVect4(gradient, shift, 1, total, 0) * ToVect4(weights, shift, 2, total * 2, 0);
   double4 temp = 1 / sqrt(ToVect4(options, shift, 3, total * 3, 1) + 1e-6);
   double4 delt = inp - ToVect4(options, shift, 3, total * 3, 0);
   double4 gvar = delt / (-2 * pow(ToVect4(options, shift, 3, total * 3, 1) + 1.0e-6, 3.0 / 2.0)) * gnx;
   double4 gmu = (-temp) * gnx - gvar * 2 * delt / (double4)batch;
   double4 gx = temp * gnx + gmu / (double4)batch + gvar * 2 * delt / (double4)batch;
//---
   D4ToArray(gradient_inputs, gx, shift, 1, total, 0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void BatchNormCalcDeltaWeights(__global double *options,
                                        __global double *delta_weights,
                                        __global double *gradients)
  {
   const int n = get_global_id(0);
   int shift_options = n * 3;
   int shift_weights = n * 2;
//---
   double grad = gradients[n];
   delta_weights[shift_weights] += grad * options[shift_options + 2];
   delta_weights[shift_weights + 1] += grad;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
__kernel void MaskMult(__global double *inputs,
                       __global double *mask,
                       __global double *outputs,
                       int outputs_total)
  {
   const int n = get_global_id(0) * 4;
//---
   double4 out = ToVect4(inputs, n, 1, outputs_total, 0) * ToVect4(mask, n, 1, outputs_total, 0);
   D4ToArray(outputs, out, n, 1, outputs_total, 0);
  }
//+------------------------------------------------------------------+
